/* adecblks.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IOMD system block level models                         ### */
/* ###                                                         ### */
/* ###  Created 17/8/92: David Flynn (as SYSBLOCKS)            ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */
/* ############################################################### */
/* ###  Revision History:                                      ### */
/* ###   18/8/92: DWF : ADEC repartitioned : Sidle/C           ### */
/* ###   25/8/92: DWF : Import Ashley's AMUX & updates to ADEC ### */
/* ###   28/8/92: DWF : partion into separate blocks           ### */
/* ###   16/9/92: AS  : stop driving data bus when not         ### */
/* ###                  accessing registers in ADEC            ### */
/* ###	 25/9/92: AS  : Added testbus output (write to ID reg) ### */
/* ###   08/1/93: WHO : fsizewt should only go if lwrite high  ### */
/* ############################################################### */


#include <Handy.h>
#include "adecblks.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}


/* ############################################################### */
/* ############################################################### */
/* ### ADEC : IOMD Address decoder                             ### */
/* ############################################################### */

#define ADECreg_def \
"ADECreg( \
  progstrb, \
  reset, \
  adec, \
  fsizewt, \
  lwrite, \
  la[], \
  d[], \
  rombank, \
  romc[], \
  fsize[],\
  testbus[]\
 )"

typedef struct
{
    ModelCore   model;
    Signal  *progstrb;
    Signal  *reset;
    Signal  *adec;
    Signal  *fsizewt;
    Signal  *lwrite;
    Bundle  *la;
    Bundle  *d;
    Signal  *rombank;
    Bundle  *romc;
    Bundle  *fsize;
    Bundle  *testbus;
} *ADECreg_model;

typedef struct
{
     int  dt;
     int  romreg0;
     int  romreg1;
     int  fsizereg;
     int  testreg[2];
} ADECreg_state;

static void ADECreg_proc (modelInst)
VoidStar modelInst;
{
   ADECreg_model me = (ADECreg_model) modelInst;
   ADECreg_state *s = (ADECreg_state *)me->model.state;
   int i, dat;

  /* async reset */
   if ( HI(reset) ) {
     s->romreg0 = s->romreg1 = 0;
     s->testreg[0] = s->testreg[1] = 0;
     BundleEvt(testbus, 
               ((s->testreg[1]<<16)+(s->testreg[0]<<8)), s->dt);     
   }

   SignalEvt(fsizewt, (HI(progstrb) && HI(lwrite) && HI(adec) && (BINVAL(la)==4)), s->dt);
   if ( HI(progstrb) && HI(adec) ) {
     if (HI(lwrite)) {
  /* register writes */
       dat=BINVAL(d);
       switch(BINVAL(la)) {
         case 0 : /*rom bank 1  reg*/
                  s->romreg0 = dat & 0x1f; break;
         case 1 : /*rom bank 1 reg*/
                  s->romreg1 = dat & 0x1f; break;
         case 4 : /*fsize register*/
                  s->fsizereg = dat & 0xff;
                  BundleEvt(fsize, s->fsizereg, s->dt);
                  break;
	 case 5 : BundleEvt(testbus, ((s->testreg[1]<<16)
                                    + (s->testreg[0]<<8) + (dat & 0xff)),
                            s->dt);
                  break;

	 case 6 :
	 case 7 : 
           s->testreg[BINVAL(la)-6] = dat & 0xff;
           BundleEvt(testbus, ((s->testreg[1]<<16)
                             + (s->testreg[0]<<8)), s->dt);
           break;
         default :  errhandler_f(0,"ADEC can't write this register\n");
       }
     } else {
  /* register reads */
/*       BundleXXX(d,s->dt); */
       switch(BINVAL(la)) {
         case 0 : /*rom bank 1  reg*/
                  dat = s->romreg0 & 0x1f; break;
         case 1 : /*rom bank 1 reg*/
                  dat = s->romreg1 & 0x1f; break;
         case 4 : /*fsize register*/
                  dat = s->fsizereg; break;
         case 5 : /*ID low byte*/
                  dat = IOMD_ID_LOW; break;
         case 6 : /*ID high byte*/
                  dat = IOMD_ID_HIGH; break;
         case 7 : /*Version number*/
                  dat = IOMD_VERSION; break;
         default :  errhandler_f(0,"ADEC whacky error!\n");
       }
       BundleEvt(d, dat, s->dt);
     }
   }
  /* disable testbus[7:0] after write */
  if (HI(adec) && FALLING(progstrb) && HI(lwrite) && (BINVAL(la) == 5))
    BundleEvt(testbus, ((s->testreg[1]<<16)+(s->testreg[0]<<8)), s->dt);
  /* disable after register reads */
   if ( HI(adec) && FALLING(progstrb) && LO(lwrite) ) {
     BundleZZZ(d,s->dt);
   }
  /* config muxes */
  /* ROM config mux */
   BundleEvt(romc, (HI(rombank) ? s->romreg1 : s->romreg0), 2*s->dt);
}

static void ADECreg_init (modelInst)
VoidStar modelInst;
{
    ADECreg_state *s;
    ADECreg_model me = (ADECreg_model) modelInst;
    NEW(s, ADECreg_state);
    me->model.proc = ADECreg_proc;
    if (!s) errhandler_f(SS_MODELS+DISASTER,"Malloc failure\n");
    s->dt = get_delay("dt",me);
    s->romreg0 = 0x1f;
    s->romreg1 = 0x1f;
    s->fsizereg = 0xff;
    me->model.state = (VoidStar ) s;
}

void ADECblks_declare()
{
    simmodel_begin("iomd.adecblks");
    simmodel_declare_s(ADECreg_def, ADECreg_init);
    simmodel_end();
}

