/* ############################################################ */
/* ###   DBUSC - Data bus control C block                   ### */
/* ###                                                      ### */
/* ###   Change history :                                   ### */
/* ###   26/10/92 : AS  : Created                           ### */

/* ############################################################ */


#include <Handy.h>
#include "dbus.h"
#include "busarb.h"

#define DBUSC_def "DBUSC(\
  {IN}  rclk,\
  {IN}  iomddec,\
  {IN}  iodec,\
  {IN}  io32dec,\
  {IN}  lwrite,\
  {IN}  dmagoing,\
  {IN}  dwrite,\
  {IN}  dmach[],\
  {IN}  dsize[],\
  {OUT} Nwbufen,\
  {OUT} Nendoutlo,\
  {OUT} Nendouthi,\
  {OUT} dbe\
)"

typedef struct {
  ModelCore model;
  Signal *rclk;
  Signal *iomddec;
  Signal *iodec;
  Signal *io32dec;
  Signal *lwrite;
  Signal *dmagoing;
  Signal *dwrite;
  Bundle *dmach;
  Bundle *dsize;
  Signal *Nwbufen;
  Signal *Nendoutlo;
  Signal *Nendouthi;
  Signal *dbe;
} ModelOfDBUSC;

typedef struct {
  int dmagoingd, dmagoingdd, dmadata, dmaactive, delay;
} DBUSCState;

static void DBUSC_proc(ModelInst)
VoidStar ModelInst;
{

  ModelOfDBUSC *me = (ModelOfDBUSC *)ModelInst;
  DBUSCState *s = me->model.state;

  if HI(rclk)  s->dmagoingd = HI(dmagoing);
  if LO(rclk) s->dmagoingdd = s->dmagoingd;

  s->dmaactive = (HI(dmagoing) && s->dmagoingd);
  s->dmadata =  (HI(dmagoing) && s->dmagoingdd);

  SignalEvt(Nendoutlo, !(
  (!s->dmaactive && LO(lwrite) && (HI(iomddec) || HI(iodec)))
  || (s->dmadata && HI(dwrite))), s->delay);

  SignalEvt(Nendouthi, !(
  (!s->dmaactive && LO(lwrite) && (HI(iomddec) || (HI(iodec) && LO(io32dec))))
  || (s->dmadata && HI(dwrite) && 
      (((BINVAL(dmach)==IODMA) && (BINVAL(dsize)!=0))
    || (BINVAL(dmach)==SND1DMA)
    || (BINVAL(dmach)==SND0DMA)))), 
  s->delay);

  SignalEvt(Nwbufen, !(
  ((HI(iomddec) || HI(iodec)) && HI(lwrite) && !s->dmaactive)
  || (s->dmadata && LO(dwrite) && 
       ((BINVAL(dmach)==IODMA)
     || (BINVAL(dmach)==SND1DMA)))),
  s->delay);

  SignalEvt(dbe, (!s->dmaactive), s->delay);

}

static void DBUSC_init(ModelInst)
VoidStar ModelInst;
{
  DBUSCState *s;
  ModelOfDBUSC *me = (ModelOfDBUSC *)ModelInst;
  NEW(s, DBUSCState);
  me->model.proc = DBUSC_proc;
  s->delay = DELAY("");
  me->model.state = (VoidStar )s;
}


void DBUSC_declare()
{
  simmodel_begin("iomd.dbusc");
  simmodel_declare_s(DBUSC_def, DBUSC_init);
  simmodel_end();
}
