/* funnel.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  FUNNEL block level model                                 ### */
/* ###                                                         ### */
/* ###  Created 14/9/92: Simon Segars                          ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */

#include <Handy.h>
#include "funnel.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}

#define FUNNEL_def \
"FUNNEL(bdin[15:0],data[31:0], fwt_lat[3:0], fwt_s[1:0], Nfrd[3:0],  frd_oe, frd_bh,\
        Nbl, Nmemc, sound) \
      => (bdout[15:0], Nblr)"

typedef struct {
    ModelCore   model;
    Bundle      *bdin, *data, *fwt_lat, *fwt_s, *Nfrd; 
    Signal      *frd_oe, *frd_bh, *Nbl, *Nmemc, *sound;
    Bundle      *bdout;
    Signal      *Nblr;
    } *FUNNEL_model;

typedef struct {
    int dt_oe, dt_gt;
    int fread0, fread1, fread2, fread3, fwrite0, fwrite1, fwrite2, fwrite3;
    } FUNNEL_state;

/*--------------------------------------------------------------------------*/

static void FUNNEL_proc (modelInst)
VoidStar modelInst;
{
    FUNNEL_model me = (FUNNEL_model) modelInst;
    FUNNEL_state *s = (FUNNEL_state *)me->model.state;
    int lat=0, dat=0;

/*----------------------------- READS ------------------------------------*/

  /* Funnel Read latches - transparent when frd_lat[n] is high */
  if HI(Nmemc)
     lat = 0xF&(~BINVAL(Nfrd));
    else
     lat = (0xF * HI(Nbl));
  dat = BINVAL(bdin);

  if ((lat & 1)==1)
     s->fread0 = dat &0xFF;

  if ((lat & 2)==2) {
     if LO(sound)
        s->fread1 = (dat>>8) &0xFF;
       else
        s->fread1 = dat &0xFF;
     }

  if ((lat & 4)==4)
     s->fread2 = dat & 0xFF;

  if ((lat & 8)==8) {
     if LO(sound)
        s->fread3 = (dat>>8) &0xFF;
       else
        s->fread3 = dat &0xFF;
     }

  SignalEvt(Nblr, lat!=0, s->dt_gt);

  /* Funnel read enable onto data bus.  All done with tristates
     words are always broadcast, even during byte and half word dmas. */

  if HI(frd_oe)
     BundleEvt(data, ((HI(frd_bh) ? s->fread3 : s->fread2)<<24 | (s->fread2)<<16 |
                      (HI(frd_bh) ? s->fread1 : s->fread0)<<8  | s->fread0), s->dt_oe);

  /* Give up the bus when I'm done with it */

  if FALLING(frd_oe)
     BundleZZZ(data, s->dt_oe);

/*-------------------------- WRITES ------------------------------------*/

  /* Write Multiplexor and latch */
  dat = BINVAL(data);
  lat = BINVAL(fwt_lat);

  if ((lat & 1)==1)
     s->fwrite0 = (dat & 0xFF);

  if ((lat & 2)==2)
     s->fwrite1 = ((dat>>8)&0xFF);

  if ((lat & 4)==4)
     s->fwrite2 = ((dat>>16)&0xFF);

  if ((lat & 8)==8)
     s->fwrite3 = ((dat>>24)&0xFF);

  switch (BINVAL(fwt_s)) {
      case 0 : BundleEvt(bdout, ((s->fwrite1<<8) | (s->fwrite0)), s->dt_oe);
               break;
      case 1 : BundleEvt(bdout, ((s->fwrite1<<8) | (s->fwrite1)), s->dt_oe);
               break;
      case 2 : BundleEvt(bdout, ((s->fwrite3<<8) | (s->fwrite2)), s->dt_oe);
               break;
      case 3 : BundleEvt(bdout, ((s->fwrite3<<8) | (s->fwrite3)), s->dt_oe);
               break;
      }

/*-------------- All done --------------------------*/
}

static void FUNNEL_init(modelInst)
VoidStar modelInst;
{
  FUNNEL_state *s;
  FUNNEL_model me = (FUNNEL_model) modelInst;
  NEW(s, FUNNEL_state);
  me->model.proc = FUNNEL_proc;
  s->dt_gt = DELAY("gt");
  s->dt_oe = DELAY("oe");

  me->model.state = (VoidStar ) s;

}

void FUNNELC_declare()
{
    simmodel_begin("iomd.funnelc");
    simmodel_declare_s(FUNNEL_def, FUNNEL_init);
    simmodel_end();
}

/*--------------------------------------------------------------------------*/
