/* iodma.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IODMA test block 				       ### */
/* ###                                                         ### */
/* ###  Created 18/11/92: Simon Segars                         ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */

#include <Handy.h>
#include "iodma.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}

#define IODMA_def \
"IODMA(bd[15:0], dack[3:0], clk8, lrnw, Nsccs, Neasis, Nior, Niorq, Niow, Nms, Nsio, Nsndcs,\
       Nscdack, Nspdack, reset, Nenbd) \
      => (drq[3:0], cdrq, id, i2cc, i2cd, Niogt, pdrq, ready, fdrq, Nbl, Nindex, Npfiq, Npirq, Nfintr, Niext,\
         Npor, Nsintr, Nscirq, pintr, flybk)"

typedef struct {
    ModelCore   model;
    Bundle      *bd, *dack;
    Signal      *clk8, *lrnw, *Nsccs, *Neasis, *Nior, *Niorq, *Niow, *Nms, *Nsio, *Nsndcs;
    Signal      *Nscdack, *Nspdack, *reset, *Nenbd;
    Bundle      *drq;
    Signal      *cdrq, *id, *i2cc, *i2cd, *Niogt, *pdrq, *ready, *fdrq, *Nbl, *Nindex, *Npfiq, *Npirq; 
    Signal      *Nfintr, *Niext, *Npor, *Nsintr, *Nscirq, *pintr, *flybk;
    } *IODMA_model;

typedef struct {
    int dt_gt;
    int drq, mode;
    int snddata, sndcnt, sioval, pcount, count;
    } IODMA_state;

/*--------------------------------------------------------------------------*/

static void IODMA_proc (modelInst)
VoidStar modelInst;
{
    IODMA_model me = (IODMA_model) modelInst;
    IODMA_state *s = (IODMA_state *)me->model.state;

if LO(reset) {
   BundleEvt(drq, 0, s->dt_gt);
   SignalEvt(ready, TRUE, s->dt_gt);
   SignalEvt(Niogt, TRUE, s->dt_gt);
   }

    if (LO(Neasis) && LO(Niow)) 
       s->mode = BINVAL(bd);

    if (HI(reset) && (BINVAL(dack)!=0xF) && RISING(Niow))
       s->drq = BINVAL(bd) & 0xFF;

    if (s->mode==1)
       BundleEvt(drq, s->drq, s->dt_gt);

    if ((s->mode&0xFF)==2)
       BundleEvt(drq, (s->mode&0xFF00)>>8, s->dt_gt);

    if HI(Nsccs)
       s->count=0;
     else if (RISING(clk8) && LO(Nior))
       s->count++;

    if (HI(reset) && (ZERO==VAL(Nsccs)) && (ZERO==VAL(Nior))) {
       BundleEvt(bd, 0xAA, s->dt_gt); /* Output AA when reading SCSI */
       SignalEvt(ready, (s->count>2), s->dt_gt);
       }

    if ((ZERO==VAL(Neasis)) && (ZERO==VAL(Nior)))
       BundleEvt(bd, 0x5A5A, s->dt_gt); /* Output 5a5a when reading EAS */

    if (RISING(Nsccs) || RISING(Neasis))
       BundleZZZ(bd, s->dt_gt);

/*-------------MEMC Podules-------------------------*/

  if HI(Nms){
     s->pcount=0;
     SignalEvt(Nbl, TRUE, s->dt_gt);
     }

  if ((ZERO==VAL(Nms)) && (ZERO==VAL(Niorq))) {
     if RISING(clk8)
        s->pcount++;
     SignalEvt(Niogt, !(s->pcount==4), s->dt_gt);
     SignalEvt(Nbl, s->pcount<2, s->dt_gt);
     if HI(lrnw){
        if (s->pcount<3)
           BundleEvt(bd, 0x1234, s->dt_gt);
          else
           BundleEvt(bd, 0x5678, s->dt_gt);
           }
     }
  if RISING(Nms)
     BundleZZZ(bd, s->dt_gt);

/*-------------IOC IO-------------------------------*/

  if LO(reset)
     s->sioval = 0xAAAA;

  if ((ZERO==VAL(Nsio)) && FALLING(Nior)) {
     BundleEvt(bd, s->sioval, s->dt_gt);
     s->sioval = s->sioval + 0x404;
     }

  if (LO(Nsio) && RISING(Nior))
     BundleZZZ(bd, 4*s->dt_gt);

/*---------------Sound CODEC-------------------------*/

  if ((ZERO==VAL(Nsndcs)) && FALLING(Nior))
     BundleEvt(bd, 0x9A56, s->dt_gt);

  if (LO(Nsndcs) && RISING(Nior))
     BundleZZZ(bd, 4*s->dt_gt);

  if (LO(Nsndcs) && LO(Niow))
     s->sndcnt = BINVAL(bd);

  if LO(reset) {
     s->sndcnt = 0;
     s->snddata = 0x9999;
     }

  if ((FALLING(Niow) || FALLING(Nior)) && LO(Nspdack))
     s->sndcnt=(s->sndcnt&0x2);

  if (FALLING(Nior) && LO(Nscdack))
     s->sndcnt=(s->sndcnt&0x1);
 
  if (FALLING(Nior) && ((ZERO==VAL(Nspdack)) || (ZERO==VAL(Nscdack)))) {
     BundleEvt(bd, s->snddata, s->dt_gt);
     s->snddata = s->snddata + 0x404;
     }

  if (RISING(Nior) && (LO(Nscdack) || LO(Nspdack)))
     BundleZZZ(bd, s->dt_gt);

  SignalEvt(pdrq, ((s->sndcnt&0x1)==0x1), s->dt_gt);

  SignalEvt(cdrq, ((s->sndcnt&0x2)==0x2), s->dt_gt);

/*-----------------Interrupts--------------------------*/

  if ((s->mode&0x0F)==4) { /* IRQA */
     SignalEvt(fdrq,   (s->mode>>8)&0x1, s->dt_gt);
     SignalEvt(Nindex, ((s->mode>>8)&0x2)==0x2, s->dt_gt);
     SignalEvt(Npfiq,  ((s->mode>>8)&0x4)==0x4, s->dt_gt);
     SignalEvt(Npirq,  ((s->mode>>8)&0x8)==0x8, s->dt_gt);
     SignalEvt(Nfintr, ((s->mode>>8)&0x10)==0x10, s->dt_gt);
     SignalEvt(Niext,  ((s->mode>>8)&0x20)==0x20, s->dt_gt);
     SignalEvt(Nsintr, ((s->mode>>8)&0x40)==0x40, s->dt_gt);
     SignalEvt(Nscirq, ((s->mode>>8)&0x80)==0x80, s->dt_gt);
     SignalEvt(pintr,  ((s->mode>>4)&0x1)==0x1, s->dt_gt);
     SignalEvt(flybk,  ((s->mode>>4)&0x2)==0x2, s->dt_gt);
     SignalEvt(Npor,   ((s->mode>>4)&0x4)!=0x4, s->dt_gt);
     }

/*-------------------I2C IOs---------------------------*/

  if ((s->mode&0x0F)==5) {
     SignalEvt(id, (s->mode>>4)&0x1, s->dt_gt);
     SignalEvt(i2cc, ((s->mode>>4)&0x2)==0x2, s->dt_gt);
     SignalEvt(i2cd, ((s->mode>>4)&0x4)==0x4, s->dt_gt);
     }

  if (RISING(Nenbd)) BundleEvt(bd, 0x0BAD, 50*100);

/*-------------- All done --------------------------*/
}

static void IODMA_init(modelInst)
VoidStar modelInst;
{
  IODMA_state *s;
  IODMA_model me = (IODMA_model) modelInst;
  NEW(s, IODMA_state);
  me->model.proc = IODMA_proc;
  s->dt_gt = DELAY("gt");

  s->drq = 0xF;
  s->mode = 0;

  me->model.state = (VoidStar ) s;

}

void IODMAC_declare()
{
    simmodel_begin("iomd.iodmac");
    simmodel_declare_s(IODMA_def, IODMA_init);
    simmodel_end();
}

/*--------------------------------------------------------------------------*/

