/* ############################################################### */
/* ############################################################### */
/* ### IOMD RAMCTLC : RAM Control signal generation            ### */
/* ###                                                         ### */
/* ### Created 28/8/92 AS                                      ### */
/* ###                                                         ### */
/* ### 02/09/92: AS: Latch address using ras rather than rclk  ### */
/* ### 03/09/92: AS: Add refresh counter and requester logic   ### */
/* ### 03/09/92: DF: change REFRQ acknowledge logic            ### */
/* ### 17/09/92: AS: Ensure Ndt and Nwe don't go off in refresh ## */
/* ### 24/9/92:  AS: Add reset state to registers              ### */
/* ###  1/2/93   WO  made refresh counter behave same as netlist # */
/* ### 17/3/93   WO  made vrefcr register with same bits as net  # */
/* ############################################################### */

#include <Handy.h>
#include "ramctl.h"
#include "busarb.h"

#define RAMCTLC_def "RAMCTLC(\
  {IN}  rclk,\
  {IN}  clk32,\
  {IN}  Ninit,\
  {IN}  dah[],\
  {IN}  dal[],\
  {IN}  ah[], \
  {IN}  al[],\
  {IN}  count0,\
  {IN}  ramreg,\
  {IN}  progstrb,\
  {IN}  la2,\
  {IN}  dwrite,\
  {IN}  lwrite,\
  {IN}  dsize[],\
  {IN}  Nbw,\
  {IN}  dmagoing,\
  {IN}  dmach[],\
  {IN}  refcres,\
  {IN}  ras,\
  {IN}  cas,\
  {IN}  rras[],\
  {IN}  vras,\
  {IN}  colmux,\
  {IN}  split,\
  {IN}  refcyc,\
  {OUT} addll[],\
  {OUT} size[],\
  {OUT} Nras[],\
  {OUT} Nvras,\
  {OUT} Nwe[],\
  {OUT} Ndt[],\
  {OUT} dsf,\
  {OUT} refrq,\
  {OUT} ramc[],\
  {IO}  d[])"

typedef struct {
  ModelCore model;
  
  Signal *rclk, *clk32, *Ninit;
  Bundle *dah, *dal, *ah, *al;
  Signal *count0, *ramreg, *progstrb, *la2, *dwrite, *lwrite;
  Bundle *dsize;
  Signal *Nbw, *dmagoing;
  Bundle *dmach;
  Signal *refcres, *ras, *cas;
  Bundle *rras;
  Signal *vras, *colmux, *split, *refcyc;
  Bundle *addll, *size, *Nras;
  Signal *Nvras;
  Bundle *Nwe, *Ndt;
  Signal *dsf, *refrq;
  Bundle *ramc;
  Bundle *d;
} ModelOfRAMCTLC;

typedef struct {

  int dt_ras, dt_cas, dt_we, dt_dt, dt_dsf, dt_regs, dt_ramc;
  int addl, addll, addh, write, addll2;
  int refcount;
  int ramcr, vrefcr;
} RAMCTLCState;

static void RAMCTLC_proc(ModelInst)
VoidStar ModelInst;
{

  ModelOfRAMCTLC *me = (ModelOfRAMCTLC *)ModelInst;
  RAMCTLCState *s = me->model.state;

/* Programmable registers */
  if LO(Ninit) {
    s->vrefcr = 0;
    s->ramcr = 0;
  } else {
    if (HI(ramreg) && HI(progstrb) && HI(lwrite)) {
      if HI(la2) 
        s->vrefcr = BINVAL(d) & 0xef;
      else
	s->ramcr = BINVAL(d) & 0xff;
    } 
  }
  if (HI(ramreg) && HI(progstrb) && LO(lwrite)) {
    if HI(la2)
      BundleEvt(d, s->vrefcr, s->dt_regs);
    else
      BundleEvt(d, s->ramcr, s->dt_regs);
  }
  if (FALLING(progstrb) && LO(lwrite)) BundleZZZ(d, s->dt_regs);

/* DMA/ARM access multiplexers */
  if LO(dmagoing) {
    if LO(ras) {
      s->addh = BINVAL(ah);
      BundleEvt(size, LO(Nbw), 5*NS);

    }
    if LO(rclk) s->addl = BINVAL(al);
    s->write = HI(lwrite);
  } else {
    if LO(ras) {
      s->addh = BINVAL(dah);
      BundleEvt(size, BINVAL(dsize), 5*NS);                   
    }
    if LO(rclk) s->addl = BINVAL(dal);
    s->write = HI(dwrite);
  }

  if LO(cas) s->addll2 = HI(count0);

  BundleEvt(ramc,
   ((s->addh & 4) ? 0x3 & (s->ramcr>>((s->addh & 3)*2))
                  : 0x7 & (s->vrefcr>>5)), s->dt_ramc);

  if LO(cas) {
    s->addll = (s->addl & 3);
    BundleEvt(addll, s->addll, 5*NS);
  }

 /* RAS generation */
  if HI(refcyc) {
    BundleEvt(Nras, ~BINVAL(rras), s->dt_ras);
    SignalEvt(Nvras, LO(vras), s->dt_ras);
  } else {
    if HI(ras) {
      if (s->addh & 4) {
        BundleEvt(Nras, ~(1<<(s->addh & 3)), s->dt_ras);
        SignalEvt(Nvras, ONE, s->dt_ras);
      } else {
        SignalEvt(Nvras, ZERO, s->dt_ras);
        BundleEvt(Nras, 0xf, s->dt_ras);
      }
    } else {
      BundleEvt(Nras, 0xf, s->dt_ras);
      SignalEvt(Nvras, ONE, s->dt_ras);
    }
  }

/* dsf, Nwe, Ndt generation */
  if HI(refcyc) {
    SignalEvt(dsf, ZERO, s->dt_dsf);
    BundleEvt(Ndt, 3, s->dt_dt);
    BundleEvt(Nwe, 3, s->dt_we);
  } else {
    if HI(colmux) { /* CAS time */
      SignalEvt(dsf, ZERO, s->dt_dsf);
      if (s->write) {
        BundleEvt(Ndt, 3, s->dt_dt);
        if ((s->addll2) && (BINVAL(ramc) & 2)) {
          BundleEvt(Nwe, 1, s->dt_we);
        } else {
          BundleEvt(Nwe, 2, s->dt_we);
        }
      } else {
        if (HI(dmagoing) && (BINVAL(dmach) == TRANDMA))
          BundleEvt(Ndt, 3, s->dt_dt);
        else {
          if ((s->addll2) && (BINVAL(ramc) & 2)) {
            BundleEvt(Ndt, 1, s->dt_dt);
          } else {
            BundleEvt(Ndt, 2, s->dt_dt);
          }
        }
        BundleEvt(Nwe, 3, s->dt_we);
      }
    } else { /* RAS time */
      if (HI(dmagoing) && (BINVAL(dmach) == TRANDMA)) {
        SignalEvt(dsf, HI(split), s->dt_dsf);
        BundleEvt(Ndt, 0, s->dt_dt);
      } else {
        SignalEvt(dsf, ZERO, s->dt_dsf);
        BundleEvt(Ndt, 3, s->dt_dt);
      }
      BundleEvt(Nwe, 3, s->dt_we);
    }
  }

/* Refresh requester counter */

  if LO(Ninit) { /* Asynchronous reset */
    s->refcount = 0;
    SignalEvt(refrq, 0, NS*7);
  } else {
    if RISING(clk32) {
      if (((s->vrefcr & 1) && ((s->refcount & 0x1FF)==0)) ||
          ((s->vrefcr & 8) && ((s->refcount & 0xFFF)==0))) {
        SignalEvt(refrq, 1, NS*7);
      }
      else if (HI(dmagoing) && (BINVAL(dmach) == REFDMA))  {
        SignalEvt(refrq, 0, NS*7);
      }
      if HI(refcres) s->refcount = 0;
      else           s->refcount = (s->refcount + 1) & 0xfff;
    }
  }
}


/*-------------- RAMCTLC Model Initialisation -----------------*/

static void RAMCTLC_init(ModelInst)
VoidStar ModelInst;
{
  RAMCTLCState *s;
  ModelOfRAMCTLC *me = (ModelOfRAMCTLC *)ModelInst;
  NEW(s, RAMCTLCState);
  me->model.proc = RAMCTLC_proc;
  s->dt_ras = DELAY("ras");
  s->dt_cas = DELAY("cas");
  s->dt_we  = DELAY("we");
  s->dt_dt  = DELAY("dt");
  s->dt_dsf = DELAY("dsf");
  s->dt_regs = DELAY("regs");
  s->dt_ramc = DELAY("ramc");
  me->model.state = (VoidStar )s;
}

void RAMCTLC_declare()
{
  simmodel_begin("iomd.ramctlc");
  simmodel_declare_s(RAMCTLC_def,      RAMCTLC_init);
  simmodel_end();
}
