/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IOMD Test blocks                                       ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */
/* ############################################################### */
/* ###  Revision History:                                      ### */
/* ###   13/10/92: AS: Created from dmatestc                   ### */
/* ############################################################### */

#include <Handy.h>
#include "testblks.h"

#define ENABLED 1
#define REQUESTING 2
#define REQ_END  4

#define NOTDEFINED 0
#define READ 1
#define WRITE 2
#define SOUND 1
#define DMA 2

/* Odd parity */
#define PARITY 1
#define DEFAULT_DIVISOR 80		/* default to 2Mhz/80 => 40us period for bit clock */

#define KBDTEST_def \
"KBDTEST(\
  {IN}  Npor,\
  {IN}  clk2,\
  {IO}  clk,\
  {IO}  data,\
  {OUT} kbstate[]\
)"

typedef struct {
  ModelCore	model;
  Signal	*Npor;
  Signal	*clk2;
  Signal	*clk;
  Signal	*data;
  Bundle	*kbdstate;
} ModelOfKBDTEST;

typedef enum {
  IDLE,		/* 0 */
  TXSTRT0,	/* 1 */
  TXSTRT1,	/* 2 */
  TXDATA0,	/* 3 */
  TXDATA1,	/* 4 */
  TXPAR0,	/* 5 */
  TXPAR1,	/* 6 */
  TXSTOP0,	/* 7 */
  TXSTOP1,	/* 8 */
  RXSDATP0,	/* 9 */
  RXSDATP1,	/* A */
  RXSTOP0,	/* B */
  RXSTOP1,	/* C */
  RXSTOPW	/* D */
} KBDState;

typedef struct {
  int count;
  KBDState state;
  int divisor;				/* ratio of Clk2 to bit clock rate */
  int every;
  int everycount;
  int start;
  int bit;
  int txdata;
  int rxdata;
  int txparity;
  int rxparity;
  int rxreg;
  int delay;
  int verbose;
} KBDTESTState;

static void KBDTEST_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfKBDTEST *me = (ModelOfKBDTEST *)ModelInst;
  KBDTESTState *s = me->model.state;
  int i;

/* Strange test to ensure clock is rising to Z1, and doesn't get clocked on
  changes from 1 to Z1 */
  if (CHANGING(clk) && HI(clk) && ZZZ(clk) && (s->state==RXSDATP0)) 
    s->rxreg = (((s->rxreg>>1) & 0xff) | (HI(data)<<8));
  
  if LO(Npor) {
    s->state = IDLE;
    s->count = 0;
    s->everycount = 0;
  } else {
    if RISING(clk2) {
      s->count = (s->count + 1) % s->divisor;
      if (s->count==0) {
      /* 40uS interval to clock the state machine */
        if (s->start) --s->start;
        if (s->start==0) ++s->everycount;

/* First decide which state to go to next */
        switch (s->state) {
    
          case IDLE :
            if ((s->everycount >= s->every) && HI(clk) && HI(data))
              s->state = TXSTRT0;
            if (HI(clk) && LO(data)) {
/* Jump to receive data - prepare for receive first */
              s->bit = 0;
              s->rxparity = 0;
              for (i=0;i<8;i++) s->rxparity = (s->rxparity ^ (s->rxdata>>i));
              s->rxparity = (s->rxparity ^ PARITY) & 1;
              s->rxreg = 0;
              s->state = RXSDATP0;
	      if (s->verbose)
		  errhandler_f (SS_MODELS+OK, "%s: starting reception at %s\n", 
				INSTNAME, bigt_tostr(&simq_now));
            }
            break;
  
          case TXSTRT0 :
	    if (LO(clk))
	    {
		if (s->verbose)
		    errhandler_f (SS_MODELS+OK, "%s: aborting transmission on start bit at %s\n",
				  INSTNAME, bigt_tostr(&simq_now));
		
		s->state = IDLE;
		s->everycount = 0;
	    }
	    else
		s->state = TXSTRT1; 
	    break;
  
          case TXSTRT1 :
            s->state = TXDATA0; break;
  
          case TXDATA0 :
            if (LO(clk))
	    {
		if (s->verbose)
		    errhandler_f (SS_MODELS+OK, "%s: aborting transmission on bit %d at %s\n",
				  INSTNAME, s->bit, bigt_tostr(&simq_now));
		    
              s->state = IDLE;
	      s->everycount = 0;
	    }
            else
              s->state = TXDATA1;
            break;
  
          case TXDATA1 :
            if (s->bit==8) s->state = TXPAR0; else s->state = TXDATA0;
            break;
  
          case TXPAR0 :
            s->state = TXPAR1; break;
  
          case TXPAR1 :
            s->state = TXSTOP0; break;

          case TXSTOP0 :
            s->state = TXSTOP1; break;
  
          case TXSTOP1 :
            s->state = IDLE; break;
    
          case RXSDATP0 :
            s->state = RXSDATP1; break;
  
          case RXSDATP1 :
            if (s->bit==10) s->state = RXSTOP0; else s->state = RXSDATP0;
            break;
  
          case RXSTOP0 :
            if (HI(data)) s->state = RXSTOP1; else s->state = RXSTOPW;
           break;
  
          case RXSTOPW :
            s->state = RXSTOP0; break;
  
          case RXSTOP1 :
	  {
	      int rdata = s->rxreg & 0xff;
	      if (rdata != s->rxdata)
	      {
		  int xpar, i;
		  /* Wrong data: but check parity on data which was actually received */
		  xpar = 0;
		  for (i = 0; i < 8; ++i)
		      xpar ^= (rdata >> i) & 1;
		  xpar ^= PARITY;	/* compute expected value of parity bit */
		  if (xpar == ((s->rxreg >> 8) & 1))
		      errhandler_f (SS_MODELS+WARNING,
				    "%s: KBD rx data error: expected %02x, got %02x at %s\n",
				    INSTNAME, s->rxdata, rdata, bigt_tostr(&simq_now));
		  else
		      errhandler_f (SS_MODELS+WARNING,
				    "%s: KBD rx data error: expected %02x, got %02x with bad parity at %s\n",
				    INSTNAME, s->rxdata, rdata, bigt_tostr(&simq_now));
	      }
	      else
	      {
		  /* Data OK: was parity correct too? */
		  if (((s->rxreg & 0x100) >> 8) != s->rxparity)
		      errhandler_f (SS_MODELS+WARNING,
				    "%s: KBD receive parity error on data %02x at %s\n",
				    INSTNAME, rdata, bigt_tostr(&simq_now));
	      }
	      s->state = IDLE;
	      s->rxdata = ((++s->rxdata) & 0xff);
	  }
            break;
  
        }
  
  /* Now output signal values to new state */
        switch (s->state) {
   
          case IDLE : 
            SignalZZZ(data, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
   
           case TXSTRT0 :
            SignalEvt(data, ZERO, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
   
          case TXSTRT1 :
            s->bit = 0;
            s->txparity = 0;
            for (i=0;i<8;i++) s->txparity = (s->txparity ^ (s->txdata>>i));
            s->txparity = (s->txparity ^ PARITY) & 1;
	    if (s->verbose)
		errhandler_f3(SS_MODELS+OK,
			      "%s: starting to xmit, data %02x\n", (int)INSTNAME, s->txdata);
            SignalEvt(data, ZERO, s->delay);
	    SignalEvt(clk, ZERO, 1*NS);
            break;
   
          case TXDATA0 :
            if ((s->txdata>>s->bit) & 1)
              SignalZZZ(data, s->delay);
            else
              SignalEvt(data, ZERO, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
   
          case TXDATA1 :
            if ((s->txdata>>s->bit) & 1)
              SignalZZZ(data, s->delay);
            else
              SignalEvt(data, ZERO, s->delay);
            SignalEvt(clk, ZERO, 1*NS);
            ++s->bit;
            break;
   
          case TXPAR0 :
            if (s->txparity)
              SignalZZZ(data, s->delay);
            else
              SignalEvt(data, ZERO, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
   
          case TXPAR1 :
            if (s->txparity)
              SignalZZZ(data, s->delay);
            else
              SignalEvt(data, ZERO, s->delay);
            if (HI(clk))		/* but if low, continue anyway! */
		SignalEvt(clk, ZERO, 1*NS);
            break;
   
          case TXSTOP0 :
            SignalZZZ(data, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
    
          case TXSTOP1 :
            s->txdata = (++s->txdata) & 0xff;
            s->everycount = 0;
            SignalZZZ(data, s->delay);
            if (HI(clk))
		SignalEvt(clk, ZERO, 1*NS);
            break;
       
          case RXSDATP0 :
            SignalZZZ(data, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
   
          case RXSDATP1 :
            ++s->bit;
            SignalZZZ(data, s->delay);
            if (HI(clk))
		SignalEvt(clk, ZERO, 1*NS);
            break;
   
          case RXSTOP0 :
            SignalZZZ(data, s->delay);
            SignalZZZ(clk, 1*NS);
            break;
   
          case RXSTOPW :
            SignalZZZ(data, s->delay);
            if (HI(clk))
		SignalEvt(clk, ZERO, 1*NS);
            break;
   
          case RXSTOP1 :
            SignalEvt(data, ZERO, s->delay);
            if (HI(clk))
		SignalEvt(clk, ZERO, 1*NS);
            break;
        }
        BundleEvt(kbdstate, s->state | s->everycount << 8, NS*5);
      }
    }
  }     
}

static void KBDTEST_init(ModelInst)
VoidStar ModelInst;
{
  KBDTESTState *s;
  ModelOfKBDTEST *me = (ModelOfKBDTEST *)ModelInst;
  Parameter param;

  NEW(s, KBDTESTState);
  me->model.proc = KBDTEST_proc;

  param = simmodel_param(me->model.state, "every");

  s->every = 0;
  if (simmodel_type(param)==PT_INTEGER) 
    s->every = simmodel_int(param);

  param = simmodel_param(me->model.state, "start");
  s->start = 0;
  if (simmodel_type(param)==PT_INTEGER)
    s->start = simmodel_int(param);

  param = simmodel_param(me->model.state, "verbose");
  s->verbose = 0;
  if (simmodel_type(param)==PT_INTEGER)
    s->verbose = simmodel_int(param);

  param = simmodel_param(me->model.state, "divisor");
  s->divisor = DEFAULT_DIVISOR;
  if (simmodel_type(param)==PT_UNDEFINED)
      s->divisor = DEFAULT_DIVISOR;
  else
  {
      if (simmodel_type(param) != PT_INTEGER)
	  errhandler_f (SS_MODELS+DISASTER,
			"%s: divisor improperly specified (wrong type)\n", INSTNAME);
      else
      {
	  int d = simmodel_int(param);
	  if (d < 2 || d > 2000)
	      errhandler_f (SS_MODELS+DISASTER,
			    "%s: divisor does not lie in range 2..2000\n", INSTNAME);
	  else
	      s->divisor = d;
      }
  }

  s->txdata = 0;
  s->rxdata = 0;
  s->delay = DELAY("");

  me->model.state = (VoidStar )s;
}

#define VIDEOTEST_def \
"VIDEOTEST(\
  {IN}  hs,\
  {IN}  vs,\
  {IN}	eclk,\
  {IN}	ed[],\
  {OUT} horiz,\
  {OUT}  vert\
)"

typedef struct {
  ModelCore	model;
  Signal	*hs;
  Signal	*vs;
  Signal	*eclk;
  Bundle	*ed;
  Signal	*horiz;
  Signal	*vert;
} ModelOfVIDEOTEST;

typedef struct {
  FILE	*f;
  int fcol;
  int HVstart;
  int HVend;
  int VVstart;
  int VVend;
  int HCount;
  int VCount;
} VIDEOTESTState;

static void VIDEOTEST_proc(ModelInst)
VoidStar ModelInst;
{
    ModelOfVIDEOTEST *me = (ModelOfVIDEOTEST *)ModelInst;
    VIDEOTESTState *s = me->model.state;

    /*
     * Model restructured by Mark Taunton, 22-Feb-93: no longer
     * attempt to synchronise data logging with VS/HS since their
     * precise relationship with the edges of ECLK is not (yet)
     * defined for VIDC20.  Instead, we log all output data (on ECLK
     * rising edge) but as an ASCII dump with unique markers for other
     * events, in particular VS/HS rise/fall.
     */
    if RISING(vs)
    {
	if (s->fcol > 0) putc ('\n', s->f);
	fputs ("V\n", s->f); fflush (s->f);
	s->VCount = 0; s->HCount = 0; s->fcol = 0;
    }	
    else if FALLING(vs)
    {
	if (s->fcol > 0) putc ('\n', s->f);
	fputs ("v\n", s->f); fflush (s->f); s->fcol = 0;
    }
    if RISING(hs)
    {
	if (s->fcol > 0) putc ('\n', s->f);
	putc ('H', s->f); fflush (s->f);
	s->HCount = 0; s->fcol = 1;
    }
    else if FALLING(hs)
    {
	if (s->fcol > 0) putc ('\n', s->f);
	putc ('h', s->f); fflush (s->f);  s->fcol = 1;
    }
    if RISING(eclk)
    {
	int data = BINVAL(ed) & 0xFF;
	++s->HCount;
	if (s->VVstart != -1)
	{
	    if ((s->VCount>=s->VVstart) && (s->VCount<s->VVend)) {
		SignalEvt(vert, ONE, 5*NS);
		if ((s->HCount>=s->HVstart) && (s->HCount<s->HVend)) {
		    SignalEvt(horiz, ONE, 5*NS);
		} else SignalEvt(horiz, ZERO, 5*NS);
	    } else SignalEvt(vert, ZERO, 5*NS);
	}
	if (s->fcol != 1)
	{
	    putc (' ', s->f);  ++s->fcol;
	}
	fprintf (s->f, "%02x", data);  s->fcol += 2;
	if (s->fcol == 96)
	{
	    putc ('\n', s->f);  s->fcol = 0;
	}
	fflush (s->f);
    }
}

static void VIDEOTEST_init(ModelInst)
VoidStar ModelInst;
{
  VIDEOTESTState *s;
  ModelOfVIDEOTEST *me = (ModelOfVIDEOTEST *)ModelInst;
  Parameter param;
  char filename[255];
  char filetype[3];
  filetype[0]=0;

  NEW(s, VIDEOTESTState);
  me->model.proc = VIDEOTEST_proc;
  param = simmodel_param(me->model.state, "filename");
  if (simmodel_type(param)!=PT_UNDEFINED) {
    strcpy(filename, simmodel_str(param));
    s->f = simfileio_find_output(NULL, filename, filetype);
    if (s->f==NULL) errhandler_f2(SS_MODELS+DISASTER,
      "%s: Couldn't open VIDEOTEST output file %s\n", (int)INSTNAME,
      (int)filename);
  } else
    errhandler_f1(SS_MODELS+DISASTER,
    "%s: VIDEOTEST output filename not specified\n",(int)INSTNAME);

  param = simmodel_param(me->model.state, "HVstart");
  if (simmodel_type(param) != PT_UNDEFINED)
  {
      if (simmodel_type(param)==PT_INTEGER)
	  s->HVstart = simmodel_int(param);
      else
	  errhandler_f1(SS_MODELS+DISASTER, 
			"%s: Horizontal video start (HVstart) improperly defined\n", (int)INSTNAME);
  }
  else
      s->HVstart = -1;
  if (simmodel_type(param) != PT_UNDEFINED)
  {
      param = simmodel_param(me->model.state, "HVend");
      if (simmodel_type(param)==PT_INTEGER)
	  s->HVend = simmodel_int(param);
      else
	  errhandler_f1(SS_MODELS+DISASTER, 
			"%s: Horizontal video end (HVend) improperly defined\n", (int)INSTNAME);
  }
  else
      s->HVend = -1;

  if (simmodel_type(param) != PT_UNDEFINED)
  {
      param = simmodel_param(me->model.state, "VVstart");
      if (simmodel_type(param)==PT_INTEGER)
	  s->VVstart = simmodel_int(param);
      else
	  errhandler_f1(SS_MODELS+DISASTER, 
			"%s: Vertical video start (VVstart) improperly defined\n", (int)INSTNAME);
  }
  else
      s->VVstart = -1;
  if (simmodel_type(param) != PT_UNDEFINED)
  {
      param = simmodel_param(me->model.state, "VVend");
      if (simmodel_type(param)==PT_INTEGER)
	  s->VVend = simmodel_int(param);
      else
	  errhandler_f1(SS_MODELS+DISASTER, 
			"%s: Vertical video end (VVend) improperly defined\n", (int)INSTNAME);
  }
  else
      s->VVend = -1;
  me->model.state = (VoidStar )s;
}

#define IORQTEST_def \
"IORQTEST(\
  {IN}  Nreset,\
  {IN}  ref8m,\
  {IN}  Niorq,\
  {IN}  Nms,\
  {IN}  lrnw,\
  {OUT} Niogt,\
  {OUT} Nbl,\
  {IO}  bd[]\
)"

typedef struct {
  ModelCore	model;
  Signal	*Nreset;
  Signal	*ref8m;
  Signal	*Niorq;
  Signal	*Nms;
  Signal	*lrnw;
  Signal	*Niogt;
  Signal	*Nbl;
  Bundle	*bd;
} ModelOfIORQTEST;

typedef struct {
  int writedata;
  int readdata;
  int delay;
  int clocks;
  int wait;
} IORQTESTState;

static void IORQTEST_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfIORQTEST *me = (ModelOfIORQTEST *)ModelInst;
  IORQTESTState *s = me->model.state;

/* Falling Niorq sets wait to zero - when times out, take Niogt low */
  if (LO(Nms) && FALLING(Niorq)) s->wait = 0;

/* Nbl driving, for writes only */
  if (LO(Nms) && LO(Niorq) && RISING(ref8m) && LO(lrnw))
    SignalEvt(Nbl, ZERO, s->delay);

  if (FALLING(ref8m) && LO(Niorq)) {
    if ((++s->wait) == s->clocks) {
      SignalEvt(Niogt, ZERO, s->delay);
      if HI(lrnw) {
        BundleEvt(bd, s->readdata, s->delay);
        s->readdata = (++s->readdata) & 0xffff;
      }
    }
  }

/* Drive Nbl for a read */
    if (LO(Niogt) && HI(lrnw) && RISING(ref8m)) 
      SignalEvt(Nbl, ZERO, s->delay);

  if (RISING(ref8m) && LO(Niorq) && LO(Niogt) && LO(lrnw) && HI(Nreset)) {
    if ((BINVAL(bd) & 0xffff) != s->writedata)
      errhandler_f3(SS_MODELS+WARNING,
      "%s: IORQTEST incorrect data. Expected %04x, got %04x\n",
      (int)INSTNAME, s->writedata, BINVAL(bd) &0xffff);
  }
  if (LO(lrnw) && LO(Niogt) && RISING(Niorq) && HI(Nreset))
    s->writedata = (++s->writedata) & 0xffff;

  if RISING(Niorq) {
    SignalEvt(Niogt, ONE, s->delay);
    BundleZZZ(bd, s->delay);
  }

  if (RISING(ref8m) && HI(Niorq)) SignalEvt(Nbl, ONE, s->delay);
}

static void IORQTEST_cmd(ModelInst, line)
VoidStar ModelInst;
char *line;
{
  ModelOfIORQTEST *me = (ModelOfIORQTEST *)ModelInst;
  IORQTESTState *s = me->model.state;
  char string[80];

  sscanf(line, "%s", string);
  if (strcmp(string, "end")==0)
    errhandler_f3(SS_MODELS+OK, "%s: readdata = %04X, writedata = %04X\n",
    (int)INSTNAME, s->readdata, s->writedata);
  else
    errhandler_f1(SS_MODELS+WARNING, "%s: Model command not understood\n",
    (int)INSTNAME);
}

static void IORQTEST_init(ModelInst)
VoidStar ModelInst;
{
  IORQTESTState *s;
  ModelOfIORQTEST *me = (ModelOfIORQTEST *)ModelInst;
  Parameter param;
  NEW(s, IORQTESTState);
  me->model.proc = IORQTEST_proc;
  me->model.cmd_proc = IORQTEST_cmd;

  param = simmodel_param(me->model.state, "wait");
  if (simmodel_type(param)==PT_INTEGER)
    s->clocks = simmodel_int(param);
  else
    s->clocks = 3;
  s->delay = DELAY("");
  param = simmodel_param(me->model.state, "start");
  if (simmodel_type(param)==PT_INTEGER)
    s->writedata = s->readdata = simmodel_int(param);
  else
    s->writedata = s->readdata = 0;
  me->model.state = (VoidStar )s;
}


#define PIOTEST_def \
"PIOTEST(\
  {IN}  Nprog,\
  {IN}  lwrite,\
  {IN}  la[],\
  {IN}  conready,\
  {IN}  contype[],\
  {IN}  clk8,\
  {IN}  clk16,\
  {IN}  Nreset,\
  {IN}  Ncs,\
  {IN}  Nior,\
  {IN}  Niow,\
  {IO}  bd[],\
  {OUT} ready,\
  {OUT} state[]\
)"

typedef struct {
  ModelCore	model;
  Signal	*Nprog;
  Signal	*lwrite;
  Bundle	*la;
  Signal	*conready;
  Bundle	*contype;
  Signal        *clk8;
  Signal	*clk16;
  Signal	*Nreset;
  Signal	*Ncs;
  Signal	*Nior;
  Signal	*Niow;
  Bundle	*bd;
  Signal	*ready;
  Bundle	*state;
} ModelOfPIOTEST;

#define S_IDLE 0
#define S_SETUP 1
#define S_STROBE 2
#define S_HOLD 3

static char *statename[4] = { "idle", "setup", "strobe", "hold" };

typedef struct
{ 
    char type;
    unsigned char setup, strobe, hold, idle_min;
} IOTime;
static IOTime io_timing[4+4] =
{
    /*
     * First 4 are new I/O space timings A..D as defined in IOMD spec,
     * measured as counts of total edges (both rising and falling) of
     * clk16.
     */
    { 'A', 3, 8, 3, 4 },			/* A */
    { 'B', 3, 6, 1, 4 },			/* B */
    { 'C', 1, 4, 1, 4 },			/* C */
    { 'D', 1, 2, 1, 4 },			/* D */
    /*
     * The next 3 are old I/O timings {slow,medium,fast,sync} and are
     * defined as numbers of rising edges of clk8 seen in the relevant
     * interval.
     */
    { '0', 2, 3, 1, 1 }, 
    { '1', 1, 3, 1, 1 },
    { '2', 1, 2, 1, 1 },
    { '3', 0, 2, 1, 3 }			/* ?? setup time ?? */
};

typedef struct {
  int readdata;
  int writedata;
  int time_errors, state_errors, data_errors;
  int delay;
  int mask;
  int width;
  int type;
  int sspace;		/* flag for S-space I/O device with address-controlled timings */
  int count;
  char *units;				/* units timing count is measured in */
  int state;
  int rdydly;
  int rdycnt;
  int genready;
  int address;
} PIOTESTState;

static void PIOTEST_proc(ModelInst)
VoidStar ModelInst;
{
    ModelOfPIOTEST *me = (ModelOfPIOTEST *)ModelInst;
    PIOTESTState *s = me->model.state;
    
    if (LO(Nreset))
    {
	SignalZZZ(ready, s->delay);
	SignalZZZ(bd, s->delay);
	s->state = S_IDLE;
	s->count = 100;			/* avoid errors on first transfer (unlikely!) */
    }
    else
    {
	/* Programming via Nprog, to enable/disable ready generation and define cycle timing */
	if (LO(Nprog) && HI(lwrite) && (BINVAL(la) == s->address))
	{
	    s->genready = HI(conready);
	    s->type = (BINVAL(contype) & 3);	/* 00=A 01=B 10=C 11=D */
	}

	if (s->sspace && RISING(clk8))
	    ++s->count;
	if (CHANGING(clk16))
	{
	    if (!s->sspace)
		++s->count;
	    if (s->state == S_STROBE && RISING (clk16) && s->genready && s->rdydly)
	    {
		++s->rdycnt;
		if (s->rdycnt == s->rdydly)
		{
		    SignalZZZ(ready, s->delay);
		    if (!s->sspace && LO(Nior))
		    {
			BundleEvt(bd, s->readdata, s->delay);
			s->readdata = (++s->readdata) & s->mask;
		    }
		}
	    }
	}


	if (FALLING(Ncs))
	{
	    IOTime *t;
	    switch (s->state)
	    {
	      case S_IDLE:
		if (s->sspace)
		{
		    /*
		     * Dynamically determine IOC-style cycle timing from
		     * latched address bus bits 20:19.
		     */
		    s->type = 4 + (BINVAL(la) >> 19 & 3);
		}
		t = &io_timing[s->type];
		if (s->count < t->idle_min)
		{
		    errhandler_f (SS_MODELS+WARNING,
				  "%s: type %c: idle time short: saw %d %s, should be >= %d, at %s\n",
				  INSTNAME, t->type, 
				  s->count, s->units, t->idle_min, bigt_tostr(&simq_now));
		    ++s->time_errors;
		}
		s->state = S_SETUP;
		s->count = 0;
		break;
	      default:
		errhandler_f (SS_MODELS+WARNING,
			      "%s: Ncs falling in wrong state (%s), at %s\n",
			      INSTNAME, statename[s->state], bigt_tostr(&simq_now));
		s->count = 0;
		++s->state_errors;
		break;
	    }
	}

	if (LO(Ncs))
	{
	    if (FALLING(Nior) || FALLING(Niow))
	    {
		IOTime *t;
		switch (s->state)
		{
		  case S_SETUP:
		    t = &io_timing[s->type];
		    if (s->count != t->setup)
		    {
			errhandler_f (SS_MODELS+WARNING,
				      "%s: type %c: setup time wrong: saw %d %s but expected %d, at %s\n",
				      INSTNAME, t->type, 
				      s->count, s->units, t->setup, bigt_tostr(&simq_now));
			++s->time_errors;
		    }
		    s->state = S_STROBE;
		    s->count = 0;
		    if (FALLING(Nior) &&
			(s->sspace || !s->genready || s->rdydly == 0))
		    {
			BundleEvt(bd, s->readdata, s->delay);
			s->readdata = (++s->readdata) & s->mask;
		    }
		    /* generate ready if so set up, even if an s-space device! */
		    if (s->genready && s->rdydly)
		    {
			SignalEvt(ready, ZERO, s->delay);
			s->rdycnt = 0;
		    }
		    break;
		    
		  default:
		    errhandler_f (SS_MODELS+WARNING,
				  "%s: Nior/Niow falling in wrong state (%s), at %s\n",
				  INSTNAME, statename[s->state], bigt_tostr(&simq_now));
		    ++s->state_errors;
		    break;
		}
	    }
	    
	    if (RISING(Nior) || RISING(Niow))
	    {
		IOTime *t;
		int time;
		switch (s->state)
		{
		  case S_STROBE:
		    t = &io_timing[s->type];
		    time = t->strobe;
		    if (!s->sspace && s->genready)
		    {
			/*
			 * Ready is supposed to stretch the strobe pulse by
			 * at least the length it was held low for.  Allow
			 * it to be longer, but by at most 2 clk16 rising
			 * edges.  Unfortunately, with the longest cycle
			 * type (A) the use of READY is not so easily
			 * accounted for in the overall timing since it is
			 * not polled for by IOMD until some way into the
			 * strobe pulse period; just ignore errors in this
			 * case.
			 */
			time += s->rdydly * 2;
			if (s->count < time && t->type != 'A')
			{
			    errhandler_f (SS_MODELS+WARNING,
					  "%s: type %c + ready: strobe too short: saw %d %s but expected >= %d, at %s\n",
					  INSTNAME, t->type, s->count, s->units, time, bigt_tostr(&simq_now));
			    ++s->time_errors;
			}
			else if (s->count > time+4)
			{
			    errhandler_f (SS_MODELS+WARNING,
					  "%s: type %c + ready: strobe too long: saw %d %s but expected <= %d, at %s\n",
					  INSTNAME, t->type, s->count, s->units, time+4, bigt_tostr(&simq_now));
			    ++s->time_errors;
			}
		    }
		    else
		    {
			if (s->count != time)
			{
			    errhandler_f (SS_MODELS+WARNING,
					  "%s: type %c: strobe wrong: saw %d %s but expected %d, at %s\n",
					  INSTNAME, t->type, s->count, s->units, time, bigt_tostr(&simq_now));
			    ++s->time_errors;
			}
		    }
		    if (RISING (Niow))
		    {
			/* Deal with writes */
			int data = BINVAL(bd) & s->mask;
			if (data != s->writedata)
			    errhandler_f(SS_MODELS+WARNING,
					 "%s: Incorrect PIO data. Expected %02x, got %02x, at %s\n",
					 INSTNAME, s->writedata, data, bigt_tostr(&simq_now));
			s->writedata = (++s->writedata) & s->mask;
		    }
		    else
		    {
			BundleZZZ(bd, s->delay);
		    }
		    s->state = S_HOLD;
		    s->count = 0;
		    break;
		    
		  default:
		    errhandler_f (SS_MODELS+WARNING,
				  "%s: Nior/Niow rising in wrong state (%s), at %s\n",
				  INSTNAME, statename[s->state], bigt_tostr(&simq_now));
		    ++s->state_errors;
		    break;
		}
	    }
	}
	if (RISING (Ncs))
	{
	    IOTime *t;
	    int time;
	    switch (s->state)
	    {
	      case S_HOLD:
		t = &io_timing[s->type];
		time = t->hold;
		if (s->count != time)
		{
		    errhandler_f (SS_MODELS+WARNING,
				  "%s: type %c: hold wrong: saw %d %s but expected %d, at %s\n",
				  INSTNAME, t->type, s->count, s->units, time, bigt_tostr(&simq_now));
		    ++s->time_errors;
		}
		s->state = S_IDLE;
		s->count = 0;
		break;
	      default:
		errhandler_f (SS_MODELS+WARNING,
			      "%s: Ncs rising in wrong state (%s), at %s\n",
			      INSTNAME, statename[s->state], bigt_tostr(&simq_now));
		++s->state_errors;
		s->count = 0;
		break;
	    }
	}
    }
    BundleEvt(state, s->state, 1);
}

static void PIOTEST_cmd(ModelInst, line)
VoidStar ModelInst;
char *line;
{
  ModelOfPIOTEST *me = (ModelOfPIOTEST *)ModelInst;
  PIOTESTState *s = me->model.state;
  char string[80];

  sscanf(line, "%s", string);
  if (strcmp(string, "end")==0)
    errhandler_f (SS_MODELS+OK,
		  "%s: readdata = %04X, writedata = %04X, data errors = %d, time errors = %d state errors = %d\n",
		  INSTNAME, s->readdata, s->writedata, s->data_errors, s->time_errors, s->state_errors);
  else
    errhandler_f1(SS_MODELS+WARNING, "%s: Model command not understood\n",
    (int)INSTNAME);
}

static void PIOTEST_init(ModelInst)
VoidStar ModelInst;
{
  PIOTESTState *s;
  ModelOfPIOTEST *me = (ModelOfPIOTEST *)ModelInst;
  Parameter param;
  int i;
  NEW(s, PIOTESTState);
  me->model.proc = PIOTEST_proc;
  me->model.cmd_proc = PIOTEST_cmd;
  param = simmodel_param(me->model.state, "width");
  /* Get width in bytes */
  s->width = 1;
  if (simmodel_type(param)==PT_INTEGER)
    s->width = simmodel_int(param);
  s->writedata = s->readdata = 0;
  s->data_errors = s->time_errors = s->state_errors = 0;
  s->mask = 0;
  for (i=0;i<s->width;i++) s->mask |= (0xff << (8*i));
  param = simmodel_param(me->model.state, "start");
  if (simmodel_type(param)==PT_INTEGER)
    s->writedata = s->readdata = simmodel_int(param);
  s->delay = DELAY("");
  param = simmodel_param(me->model.state, "sspace");
  if (simmodel_type == PT_UNDEFINED)
      s->sspace = 0;
  else
      s->sspace = simmodel_int(param);
  if (s->sspace)
      s->units = "rising edges of clk8";
  else
      s->units = "half-ticks of clk16";
  param = simmodel_param(me->model.state, "ready");
  if (simmodel_type == PT_UNDEFINED)
      s->rdydly = 0;
  else
      s->rdydly = simmodel_int(param);
  param = simmodel_param(me->model.state, "address");
  if (simmodel_type(param)!=PT_INTEGER)
    errhandler_f1(SS_MODELS+DISASTER, "%s: No address for PIOTEST given\n",
    (int)INSTNAME);
  else
    s->address = simmodel_int(param);

  me->model.state = (VoidStar )s;
}


#define DMATEST_def \
"DMATEST(\
  {IN}	Nprog,\
  {IN}  lwrite,\
  {IN}  la[],\
  {IO}  d[],\
  {IN}  clk16,\
  {IN}  Ndack,\
  {IN}  Niow,\
  {IN}  Nior,\
  {IN}	tc,\
  {OUT} drq,\
  {IO}  bd[]\
)"

typedef struct {
  ModelCore	model;
  Signal	*Nprog;
  Signal	*lwrite;
  Bundle	*la;
  Bundle	*d;
  Signal	*clk16;
  Signal	*Ndack;
  Signal	*Niow;
  Signal	*Nior;
  Signal	*tc;
  Signal	*drq;
  Bundle	*bd;
} ModelOfDMATEST;

typedef struct {
  int delay;
  int address;
  int every;
  int type;
  int bytes;
  int sbyte;
  int width;
  int count;
  int status;
  int verbose;
  int drq_hold;
  int expected;
  int mode;
  int started;
  int data;
  int mask;
  int transfers;
  int blocks;
  int data_errors;
} DMATESTState;

static void DMATEST_proc(ModelInst)
VoidStar ModelInst;
{
    ModelOfDMATEST *me = (ModelOfDMATEST *)ModelInst;
    DMATESTState *s = me->model.state;

    if (!s->started)
    {
	SignalEvt(drq, ZERO, s->delay);
	s->started = TRUE;
    }

    /* Programming via Nprog, to enable/disable DMA requests */
    if (HI(lwrite) && RISING(Nprog) && BINVAL(la) == s->address)
    {
	int data = BINVAL(d);
	/* Low bit of data controls enabled bit */
	if (data & 1)
	    s->status |= ENABLED;
	else
	    s->status &= ~ENABLED;
	s->bytes = (data >> 1) & 3;		/* Bits 2:1 are sound bytes/request MOD 4 */
	if (s->bytes == 0)
	    s->bytes = 4;
	
	if (!(s->status & ENABLED))
	{
	    if (s->status & REQUESTING)
	    {
		if (LO(Ndack) || LO(Nior) || LO(Niow))
		{
		    errhandler_f (SS_MODELS+ERROR,
				  "%s: DMA disabled while request being serviced (software error?) at %s\n",
				  bigt_tostr(&simq_now));
		    if (LO(Nior))
			BundleZZZ(bd, 1); /* tristate with minimal delay */
		}
		SignalEvt(drq, ZERO, s->delay); /* take it away with low delay */
	    }
	    if (s->verbose)
		errhandler_f (SS_MODELS+OK, "%s: DMA disabled%s at %s\n",
			      INSTNAME,
			      s->status & REQUESTING ? " (outstanding request removed)" : "",
			      bigt_tostr (&simq_now));
	    s->status &= ~(REQUESTING|REQ_END);
	}
	else
	{
	    s->blocks++;		/* increment count of blocks set up */
	    s->transfers = 0;		/* transfers measured per block */
	    s->expected = 0;		/* reset data value sequence */
	    s->count = 0;		/* reset inter-request counter now */
	    /*
	     * Mark Taunton, 24-Feb-93: add programming of DMA
	     * inter-request interval for approximation to some of the
	     * rates supported by the AD1848 and CS4248 codecs (though not
	     * encoded the same way!).  Uses data bits 5:3 to encode
	     * requested interval: 0 = no change, 1..7 give specific
	     * periods.
	     */
	    if (s->type != SOUND) 
	    {
		if (s->verbose)
		    errhandler_f (SS_MODELS+OK,
				  "%s: I/O DMA set up, block %d at %s\n",
				  INSTNAME, s->blocks, bigt_tostr (&simq_now));
	    }
	    else
	    {
		if (s->verbose)
		    errhandler_f (SS_MODELS+OK,
				  "%s: sound DMA set up, block %d: speed %d bytes %d at %s\n",
				  INSTNAME, s->blocks, (data >> 3) & 7, s->bytes, bigt_tostr (&simq_now));
		switch ((data >> 3) & 7)
		{
		  case 0:				/* no change */
		    break;
		  case 1:
		    s->every = 16*1000000/8000;	/* 8k (Telephony) */
		    break;
		  case 2:
		    s->every = 16*1000000/11025;	/* 11k025 (CD/4) */
		    break;
		  case 3:
		    s->every = 16*1000000/22050;	/* 22k05 (CD/2) */
		    break;
		  case 4:
		    s->every = 16*1000000/32000;	/* 32k (NICAM, LP-DAT) */
		    break;
		  case 5:
		    s->every = 16*1000000/37800;	/* 37k8 (CD-XA dig audio, high rate) */
		    break;
		  case 6:
		    s->every = 16*1000000/44100;	/* 44k1 (CD) */
		    break;
		  case 7:
		    s->every = 16*1000000/48000;	/* 48k (DAT) */
		    break;
		  default:				/* eh? */
		    break;
		}
	    }
	}
    }
    if (RISING(clk16) && (s->status & ENABLED)) 
    {
	/* Set drq high if enabled and it's time to do so */
	if (!(s->status & (REQUESTING|REQ_END)))
	{
	    ++s->count;
	    if (s->count >= s->every) 
	    {
		SignalEvt(drq, ONE, s->delay);
		s->status |= REQUESTING; /* Set requesting flag */
		s->sbyte = 0;		/* reset byte counter (for sound) */
		s->count = 0;
		s->transfers++;
	    }
	}
	else
	{
	    if (s->type == SOUND) 
	    {
		/*
		 * The real sound device generates/consumes data at fixed
		 * rate, not at a fixed time after previous request was
		 * satisfied.  So we keep the count going and check that the
		 * request was serviced in time.  On failure, the chip would
		 * just fill-in for the sample on playback, or discard the
		 * following sample on capture, and wait till the DMA
		 * finished.
		 */
		++s->count;
		if (s->count % s->every == 0) 
		{
		    if (s->count == s->every) /* first time */
			errhandler_f (SS_MODELS+WARNING, 
				      "%s: sound DMA delayed too long on block %d transfer %d at %s\n",
				      INSTNAME, s->blocks, s->transfers, bigt_tostr (&simq_now));
		    else if (s->count % s->every == 0 && s->verbose > 1)
			errhandler_f (SS_MODELS+WARNING, 
				      "%s: sound DMA delayed %d samples too long: block %d transfer %d at %s\n",
				      INSTNAME, s->count / s->every, s->blocks, s->transfers,
				      bigt_tostr (&simq_now));
		}
	    }
	}
    }
    
    if ((s->status & (REQUESTING|REQ_END)) && LO(Ndack)) 
    {
	
	/* Disable channel requests if TC activated */
	if (HI(tc))
	{
	    if (s->status & ENABLED)
		if (s->verbose)
		    errhandler_f (SS_MODELS+OK,
				  "%s: TC asserted on block %d, transfer %d at %s\n",
				  INSTNAME, s->blocks, s->transfers, bigt_tostr(&simq_now));
	    s->status &= ~ENABLED;
	}
	
	/* Calculate data for this transfer */
	if (FALLING(Nior) || FALLING(Niow)) 
	{
	    int i;
	    s->data = 0;
	    for (i = 0; i < s->width; i++) 
	    {
		s->data |= (s->expected & 0xFF) << (8*i);
		s->expected++;
	    }
	}

	/* Check data written to DMATEST on rising Niow */
	if (s->mode == WRITE && RISING(Niow))
	{
	    if ((s->data & s->mask) != (BINVAL(bd) & s->mask))
	    {
		errhandler_f (SS_MODELS+WARNING,
			      "%s: Incorrect DMA data. Expected %04x, got %04x at %s\n", (int)INSTNAME,
			      (s->data & s->mask), (BINVAL(bd) & s->mask), bigt_tostr(&simq_now));
		++s->data_errors;
	    }
	}

	/* Output data for read from DMATEST on falling Nior */
	if (s->mode == READ && FALLING(Nior)) 
	    BundleEvt(bd, s->data, s->delay);

	/* Disable drq if end of transfer */
	if (FALLING(Nior) || FALLING(Niow))
	{
	    ++s->sbyte;
	    if (s->type != SOUND || (s->type == SOUND && s->sbyte == s->bytes))
	    {
		SignalEvt(drq, ZERO, s->drq_hold);
		s->status &= ~REQUESTING;
		s->status |= REQ_END;
	    }
	}

	/* Take data away at end of read strobe */
	if (s->mode == READ && RISING(Nior))
	    BundleZZZ(bd, s->delay);

	/* At end of (last) strobe, clear down internal activity flag */
	if ((s->status & REQ_END) && (RISING(Nior) || RISING(Niow)))
	{
	    s->status &= ~REQ_END;
	    if (s->verbose > 1)
		errhandler_f (SS_MODELS+OK,
			      "%s: end of DMA block %d transfer %d at %s\n",
			      INSTNAME, s->blocks, s->transfers, bigt_tostr(&simq_now));
	}
    }
}

static void DMATEST_cmd(ModelInst, line)
VoidStar ModelInst;
char *line;
{
  ModelOfDMATEST *me = (ModelOfDMATEST *)ModelInst;
  DMATESTState *s = me->model.state;
  char string[80];

  sscanf(line, "%s", string);
  if (strcmp(string, "end")==0)
  {
      if (s->mode == READ)
	  errhandler_f (SS_MODELS+OK, "%s: data reached %04X\n",
			(int)INSTNAME, s->expected);
      else
	  errhandler_f (SS_MODELS+OK, "%s: data reached %04X, data errors = %d\n",
			(int)INSTNAME, s->expected, s->data_errors);
  }      
  else
      errhandler_f (SS_MODELS+WARNING, "%s: Model command not understood\n", (int)INSTNAME);
}

static void DMATEST_init(ModelInst)
VoidStar ModelInst;
{
  DMATESTState *s;
  ModelOfDMATEST *me = (ModelOfDMATEST *)ModelInst;
  Parameter param;
  NEW(s, DMATESTState);
  me->model.proc = DMATEST_proc;
  me->model.cmd_proc = DMATEST_cmd;
  param = simmodel_param(me->model.state, "address");
  if (simmodel_type(param) != PT_INTEGER) errhandler_f1(SS_MODELS+DISASTER,
  "%s: Address not specified/incorrectly specified\n", (int)INSTNAME);
  else s->address = simmodel_int(param);

  param = simmodel_param(me->model.state, "width");
  if (simmodel_type(param) != PT_INTEGER) 
      errhandler_f1(SS_MODELS+DISASTER,
		    "%s: Bus width not defined/incorrectly specified\n", (int) INSTNAME);
  else 
      s->width = simmodel_int(param);
  if (s->width == 1)
      s->mask = 0xFF;
  else if (s->width == 2)
      s->mask = 0xFFFF;
  else if (s->width == 4)
      s->mask = 0xFFFFFFFF;
  else
      errhandler_f (SS_MODELS+DISASTER,
		    "%s: Width must be 1, 2, or 4 bytes", INSTNAME);  
  
  param = simmodel_param(me->model.state, "type");
  if (simmodel_type(param) != PT_STRING) 
      errhandler_f (SS_MODELS+DISASTER,
		    "%s: DMA type not defined/wrongly specified\n", INSTNAME);
  if (cistrcmp(param, "sound"))
      s->type = SOUND;
  else if (cistrcmp(param, "dma")) 
      s->type = DMA;
  else 
      errhandler_f (SS_MODELS+DISASTER,
		    "%s: DMA Type must be either dma or sound\n", INSTNAME);

  /* Get default number of sound bytes (changeable by programming) */
  if (s->type == SOUND)
  {
    s->bytes = 1;
    param = simmodel_param(me->model.state, "bytes");
    if (simmodel_type(param) != PT_UNDEFINED) 
    {
	if (simmodel_type(param) == PT_INTEGER)
	    s->bytes = simmodel_int(param);
	else
	    errhandler_f (SS_MODELS+DISASTER, 
			  "%s: Sound DMA bytes parameter not correctly specified\n", INSTNAME);
    }
  }

  param = simmodel_param(me->model.state, "mode");
  if (simmodel_type(param) != PT_STRING) 
      errhandler_f (SS_MODELS+DISASTER,
		    "%s: DMA mode not correctly specified\n", (int)INSTNAME);
  if (cistrcmp(param, "write")) 
      s->mode = WRITE;
  else if (cistrcmp(param, "read")) 
      s->mode = READ;
  else
      errhandler_f (SS_MODELS+DISASTER,
		    "%s: DMA Mode must be either read or write\n", INSTNAME);

  param = simmodel_param(me->model.state, "every");
  if (simmodel_type(param) != PT_INTEGER) 
      errhandler_f (SS_MODELS+DISASTER, 
		    "%s: DMA every parameter not correctly specified\n", INSTNAME);
  s->every = simmodel_int(param);

  param = simmodel_param(me->model.state, "verbose");
  if (simmodel_type(param) == PT_UNDEFINED)
      s->verbose = 0;
  else
  {
      if (simmodel_type(param) != PT_INTEGER) 
      {
	  errhandler_f (SS_MODELS+WARNING, 
			"%s: DMA verbose parameter not correctly specified\n", INSTNAME);
	  s->verbose = 1;
      }
      else
	  s->verbose = simmodel_int(param);
  }  
  s->status = 0;
  s->count = 0;
  s->delay = DELAY("dt");
  s->drq_hold = DELAY("drq");
  s->expected = 0;
  s->sbyte = 0;
  s->started = FALSE;
  s->transfers = 0;  s->blocks = 0;
  s->data_errors = 0;
  me->model.state = (VoidStar )s;
}


#define IO_TIMING_def \
"IO_TIMING(\
  {IN}  Nprog,\
  {IN}  lrnw,\
  {IN}  la[],\
  {IN}  d[],\
  {IN}  clk16,\
  {IN}  Nselect[],\
  {IN}  Niow,\
  {IN}  Nior\
)"

typedef struct {
  ModelCore	model;
  Signal	*Nprog;
  Signal	*lrnw;
  Bundle	*la;
  Bundle	*d;
  Signal	*clk16;
  Bundle	*Nselect;
  Signal	*Niow;
  Signal	*Nior;
} ModelOfIO_TIMING;

#define MAXSELECTS 16
typedef struct {
  int timing[MAXSELECTS], bytes[MAXSELECTS];
  int selects, selmask;
  int channel; /* Currently active channel */
  int status, bytecount, count;		/* info on current transaction */
  int address;
} IO_TIMINGState;

#define STRWIDTH_MASK	0xf
#define STRWIDTH_SHIFT	0
#define SELSETUP_MASK	0xf
#define SELSETUP_SHIFT	4
#define SELHOLD_MASK	0xf
#define SELHOLD_SHIFT	8

#if 0
#define DRQHOLD  0x0000f000
#define DRQEVERY 0x7fff0000
#define REQUEST  0x80000000
#endif

#define IDLE 0
#define SELECT_SETUP 1
#define STROBE 2
#define SELECT_HOLD 3

static void IO_TIMING_proc(ModelInst)
VoidStar ModelInst;
{
    ModelOfIO_TIMING *me = (ModelOfIO_TIMING *)ModelInst;
    IO_TIMINGState *s = me->model.state;

    int i;

    /* Allow programming of required timing */
    if (((BINVAL(la) & 0xffffffC0) == s->address) && LO(Nprog) && LO(lrnw))
	s->timing[BINVAL(la) >> 2 & 0xF] = BINVAL(d);
    /* Allow programming of number of bytes per transfer */
    if (((BINVAL(la) & 0xffffffC0) == (s->address+0x40)) && LO(Nprog) && LO(lrnw))
	s->bytes[BINVAL(la) >> 2 & 0xF] = BINVAL(d);

    if CHANGING(Nselect) 
    {
	int new = (~BINVAL(Nselect) & s->selmask); /* NB new is bitset in +ve logic */
	if (new == 0)
	{
	    /* None now active: check timings on the last which was */
	    if (s->channel == -1)
		errhandler_f(SS_MODELS+WARNING,
			     "%s: Selects changed but none were or are active at %s\n",
			     INSTNAME, bigt_tostr(&simq_now));
	    else
	    {
		int bytec, bytes, timing;
		switch (s->status)
		{
		  case SELECT_HOLD:	/* OK */
		    if ((bytec = s->bytecount) != (bytes = s->bytes[s->channel]))
			errhandler_f(SS_MODELS+WARNING, 
				     "%s: Incorrect number of bytes (%d should be %d) transferred on channel %d at %s\n",
				     INSTNAME, s->channel, bytec, bytes, bigt_tostr(&simq_now));
		    if (s->count != (timing = (s->timing[s->channel] >> SELHOLD_SHIFT & SELHOLD_MASK)))
			errhandler_f2(SS_MODELS+WARNING,
				      "%s: Incorrect select hold time (%d should be %d) on channel %d at %s\n", 
				      (int)INSTNAME, s->count, timing, s->channel, bigt_tostr(&simq_now));
		    break;
		  case SELECT_SETUP:
		    errhandler_f(SS_MODELS+WARNING,
				 "%s: Select removed before r/w strobe on channel %d at %s\n",
				 INSTNAME, s->channel, bigt_tostr(&simq_now));
		    break;
		  case STROBE:
		    errhandler_f(SS_MODELS+WARNING,
				 "%s: Select removed during strobe pulse on channel %d at %s\n",
				 INSTNAME, s->channel, bigt_tostr(&simq_now));
		    break;
		  default:
		    errhandler_f(SS_MODELS+WARNING,
				 "%s: Select in bad state (%d) for channel %d at %s\n",
				 INSTNAME, s->status, s->channel, bigt_tostr(&simq_now));
		    break;
		}
		s->channel = -1;
		s->status = IDLE;
		s->count = 0;
		s->bytecount = 0;
	    }
	}
	else
	{
	    int chan;
	    /* Check for only one select being active: use cute check for new being 2**N */
	    if ((new & -new) != new)
	    {
		errhandler_f(SS_MODELS+WARNING, 
			     "%s: multiple selects active (0x%4x) at %s\n",
			     INSTNAME, new, bigt_tostr(&simq_now));
		if (s->channel == -1)
		    /* No channel was active: will find lowest numbered of new ones below */
		    ;
		else
		    /* Some channel was already active: ignore the later one! */
		    new = 1 << s->channel;
	    }
	    /* Find channel that has just selected */
	    for (chan = 0; chan < s->selects && !(new & (1 << chan)); ++chan)
		;
	    s->channel = chan;
	    s->status = SELECT_SETUP;
	    s->count = 0;	/* start timing select setup phase */
	    s->bytecount = 0;
	}
    }

    if (CHANGING(clk16))
	++s->count;
    
    if (FALLING(Nior) || FALLING(Niow)) 
    {
	if (s->channel == -1)
	    errhandler_f(SS_MODELS+WARNING,
			 "%s: Nior/Niow activated while no device selected at %s\n",
			 INSTNAME, bigt_tostr(&simq_now));
	else
	{
	    int timing;
	    if (s->status != SELECT_SETUP)
		errhandler_f(SS_MODELS+WARNING,
			     "%s: Nior/Niow falling in wrong state (%d) on channel %d at %s\n",
			     INSTNAME, s->status, s->channel, bigt_tostr(&simq_now));
	    if (s->count != (timing = (s->timing[s->channel] >> SELSETUP_SHIFT & SELSETUP_MASK)))
		errhandler_f(SS_MODELS+WARNING,
			     "%s: Incorrect select setup time (%d should be %d) on channel %d at %s\n",
			     INSTNAME, s->count, timing, s->channel, bigt_tostr(&simq_now));
	    s->status = STROBE;
	}
	s->count = 0;
	++s->bytecount;
    }
    
    if (RISING(Nior) || RISING(Niow)) 
    {
	if (s->channel == -1)
	{
	    /* Will already have reported a message about this */
#if 0
	    errhandler_f(SS_MODELS+WARNING,
			 "%s: Nior/Niow going away while no device selected at %s\n",
			 INSTNAME, bigt_tostr(&simq_now));
#endif
	}
	else
	{
	    int timing;
	    if (s->status != STROBE) 
		errhandler_f(SS_MODELS+WARNING,
			     "%s: Nior/Niow rising when not in STROBE state on channel %d at %s\n",
			     INSTNAME, s->channel, bigt_tostr(&simq_now));
	    if (s->count != (timing = (s->timing[s->channel] >> STRWIDTH_SHIFT & STRWIDTH_MASK)))
		errhandler_f(SS_MODELS+WARNING, 
			     "%s: Incorrect Nior/Niow width (%d should be %d) for channel %d at %s\n",
			     INSTNAME, s->count, timing, s->channel, bigt_tostr(&simq_now));
	    if (s->bytecount < s->bytes[s->channel])
		s->status = SELECT_SETUP;
	    else
		s->status = SELECT_HOLD;
	}
    }
}

static void IO_TIMING_init(ModelInst)
VoidStar ModelInst;
{
  IO_TIMINGState *s;
  int i;
  Parameter param;
  ModelOfIO_TIMING *me = (ModelOfIO_TIMING *)ModelInst;

  NEW(s, IO_TIMINGState);
  me->model.proc = IO_TIMING_proc;
  param = simmodel_param(me->model.state, "address");
  if (simmodel_type(param)!=PT_INTEGER)
    errhandler_f1(SS_MODELS+DISASTER,
    "%s: Address not specified", (int)INSTNAME);
  s->address = simmodel_int(param);
  /* Get number of select lines from width of supplied Nselect[] bundle: check manageable */
  s->selects = (me->Nselect->flags&LENGTH_BITS) + 1;
  if (s->selects > MAXSELECTS)
      errhandler_f(SS_MODELS+DISASTER,
		   "%s: too many select lines to handle (%d, max %d)\n", 
		   INSTNAME, s->selects, MAXSELECTS);
  /* Create mask for bundle of select lines. */
  s->selmask = ~(~0 << s->selects);
  /* Default to 1 byte transfers for each channel */
  for (i = 0; i < s->selects; i++) 
      s->bytes[i] = 1;
  s->status = IDLE;
  s->channel = -1;
  me->model.state = (VoidStar )s;
}

#define MOUSETEST_def "MOUSETEST(\
 Xlrnw,\
 XNprog,\
 Xla[],\
 Xd[],\
 Xin[],\
 Xout[])"

typedef struct {
  ModelCore model;
  Signal *Xlrnw;
  Signal *XNprog;
  Bundle *Xla;
  Bundle *Xd;
  Bundle *Xin;
  Bundle *Xout;
} ModelOfMOUSETEST;

typedef struct {
  int out_delay;
  int in_delay;
  int address;
} MOUSETESTState;

static void MOUSETEST_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfMOUSETEST *me = (ModelOfMOUSETEST *)ModelInst;
  MOUSETESTState *s = me->model.state;

  int dat;

  if (LO(XNprog) && LO(Xlrnw) && (s->address == BINVAL(Xla))) {
    dat = BINVAL(Xd);
    BundleEvt(Xout, dat, s->out_delay);
   }
  if (FALLING(XNprog) && HI(Xlrnw) && (s->address == BINVAL(Xla))) {
    dat = BINVAL(Xin);
    BundleEvt(Xd, dat, s->in_delay);
  }
  if (RISING(XNprog) && HI(Xlrnw) && (s->address == BINVAL(Xla)))
    BundleZZZ(Xd, s->in_delay);
}

static void MOUSETEST_init(ModelInst)
VoidStar ModelInst;
{
  MOUSETESTState *s;
  Parameter param;
  ModelOfMOUSETEST *me = (ModelOfMOUSETEST *)ModelInst;
  NEW(s, MOUSETESTState);
  me->model.proc = MOUSETEST_proc;
  s->out_delay = DELAY("out");
  s->in_delay = DELAY("in");
  param = simmodel_param(me->model.state, "address");
  if (simmodel_type(param)!=PT_INTEGER)
    errhandler_f1(SS_MODELS+DISASTER,
    "%s: Address not specified", (int)INSTNAME);
  s->address = simmodel_int(param);
  me->model.state = (VoidStar )s;
}

void TESTBlks_declare()
{
  simmodel_begin("iomd.testblks");
  simmodel_declare_s(KBDTEST_def, KBDTEST_init);
  simmodel_declare_s(VIDEOTEST_def, VIDEOTEST_init);
  simmodel_declare_s(IORQTEST_def, IORQTEST_init);
  simmodel_declare_s(PIOTEST_def, PIOTEST_init);
  simmodel_declare_s(DMATEST_def, DMATEST_init);
  simmodel_declare_s(IO_TIMING_def, IO_TIMING_init);
  simmodel_declare_s(MOUSETEST_def, MOUSETEST_init);
  simmodel_end();
}
